"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = exports.PgMetaschemaPlugin = void 0;

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _taggedTemplateLiteral2 = _interopRequireDefault(require("@babel/runtime/helpers/taggedTemplateLiteral"));

var _graphileUtils = require("graphile-utils");

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { (0, _defineProperty2["default"])(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _templateObject() {
  var data = (0, _taggedTemplateLiteral2["default"])(["\n        type MetaschemaType {\n          pgAlias: String!\n          pgType: String!\n          gqlType: String!\n          subtype: String\n          modifier: Int\n          typmod: JSON\n          isArray: Boolean!\n        }\n        type MetaschemaField {\n          name: String!\n          type: MetaschemaType!\n        }\n        type MetaschemaTableInflection {\n          # https://github.com/graphile/graphile-engine/blob/v4/packages/graphile-build-pg/src/plugins/PgBasicsPlugin.js\n          allRows: String!\n          allRowsSimple: String!\n          tableFieldName: String!\n          tableType: String!\n          createPayloadType: String!\n          orderByType: String!\n          filterType: String\n          inputType: String!\n          patchType: String\n          conditionType: String!\n          patchField: String!\n          edge: String!\n          edgeField: String!\n          connection: String!\n          typeName: String!\n          enumType: String!\n\n          updatePayloadType: String\n          deletePayloadType: String!\n          deleteByPrimaryKey: String\n          updateByPrimaryKey: String\n\n          createField: String!\n          createInputType: String!\n        }\n        type MetaschemaTableQuery {\n          all: String!\n          one: String!\n          create: String!\n          update: String\n          delete: String\n        }\n        type MetaschemaTable {\n          name: String!\n          query: MetaschemaTableQuery!\n          inflection: MetaschemaTableInflection!\n          fields: [MetaschemaField]\n          constraints: [MetaschemaConstraint]\n          foreignKeyConstraints: [MetaschemaForeignKeyConstraint]\n          primaryKeyConstraints: [MetaschemaPrimaryKeyConstraint]\n          uniqueConstraints: [MetaschemaUniqueConstraint]\n          checkConstraints: [MetaschemaCheckConstraint]\n          exclusionConstraints: [MetaschemaExclusionConstraint]\n        }\n        union MetaschemaConstraint =\n            MetaschemaForeignKeyConstraint\n          | MetaschemaUniqueConstraint\n          | MetaschemaPrimaryKeyConstraint\n          | MetaschemaCheckConstraint\n          | MetaschemaExclusionConstraint\n        type MetaschemaForeignKeyConstraint {\n          name: String!\n          fields: [MetaschemaField]\n          refTable: MetaschemaTable\n          refFields: [MetaschemaField]\n        }\n        type MetaschemaUniqueConstraint {\n          name: String!\n          fields: [MetaschemaField]\n        }\n        type MetaschemaPrimaryKeyConstraint {\n          name: String!\n          fields: [MetaschemaField]\n        }\n        type MetaschemaCheckConstraint {\n          name: String!\n          fields: [MetaschemaField]\n        }\n        type MetaschemaExclusionConstraint {\n          name: String!\n          fields: [MetaschemaField]\n        }\n        type Metaschema {\n          tables: [MetaschemaTable]\n        }\n        extend type Query {\n          _meta: Metaschema\n        }\n      "]);

  _templateObject = function _templateObject() {
    return data;
  };

  return data;
}

var GIS_TYPES = ['Geometry', 'Point', 'LineString', 'Polygon', 'MultiPoint', 'MultiLineString', 'MultiPolygon', 'GeometryCollection'];

var aliasTypes = function aliasTypes(type) {
  switch (type) {
    case 'int8':
      return 'bigint';

    case 'bool':
      return 'boolean';

    case 'bpchar':
      return 'char';

    case 'float8':
      return 'float';

    case 'float4':
      return 'real';

    case 'int4':
      return 'int';

    case 'int2':
      return 'smallint';

    default:
      return type;
  }
};

var PgMetaschemaPlugin = (0, _graphileUtils.makeExtendSchemaPlugin)(function (build, schemaOptions) {
  /** @type {import('graphile-build-pg').PgIntrospectionResultsByKind} */
  var introspection = build.pgIntrospectionResultsByKind;
  var inflection = build.inflection;
  /** @type {string[]} */

  var schemas = schemaOptions.pgSchemas;
  var pgGetGqlTypeByTypeIdAndModifier = build.pgGetGqlTypeByTypeIdAndModifier;
  return {
    typeDefs: (0, _graphileUtils.gql)(_templateObject()),
    resolvers: {
      // TODO determine why check constraints aren't coming through
      MetaschemaCheckConstraint: {
        /** @param constraint {import('graphile-build-pg').PgConstraint} */
        fields: function fields(constraint) {
          return constraint.keyAttributes;
        }
      },
      MetaschemaExclusionConstraint: {
        /** @param constraint {import('graphile-build-pg').PgConstraint} */
        fields: function fields(constraint) {
          return constraint.keyAttributes;
        }
      },
      MetaschemaUniqueConstraint: {
        /** @param constraint {import('graphile-build-pg').PgConstraint} */
        fields: function fields(constraint) {
          return constraint.keyAttributes;
        }
      },
      MetaschemaPrimaryKeyConstraint: {
        /** @param constraint {import('graphile-build-pg').PgConstraint} */
        fields: function fields(constraint) {
          return constraint.keyAttributes;
        }
      },
      MetaschemaForeignKeyConstraint: {
        /** @param constraint {import('graphile-build-pg').PgConstraint} */
        fields: function fields(constraint) {
          return constraint.keyAttributes;
        },

        /** @param constraint {import('graphile-build-pg').PgConstraint} */
        refTable: function refTable(constraint) {
          return constraint.foreignClass;
        },

        /** @param constraint {import('graphile-build-pg').PgConstraint} */
        refFields: function refFields(constraint) {
          return constraint.foreignKeyAttributes;
        }
      },
      MetaschemaType: {
        /** @param attr {import('graphile-build-pg').PgType} */
        pgType: function pgType(type) {
          var _type$arrayItemType;

          // TODO what is the best API here?
          // 1. we could return original _name, e.g. _citext (= citext[])
          // 2. we could return original type name and include isArray
          if (type.isPgArray && ((_type$arrayItemType = type.arrayItemType) === null || _type$arrayItemType === void 0 ? void 0 : _type$arrayItemType.name)) {
            return type.arrayItemType.name;
          }

          return type.name;
        },
        pgAlias: function pgAlias(type) {
          var _type$arrayItemType2;

          if (type.isPgArray && ((_type$arrayItemType2 = type.arrayItemType) === null || _type$arrayItemType2 === void 0 ? void 0 : _type$arrayItemType2.name)) {
            return aliasTypes(type.arrayItemType.name);
          }

          return aliasTypes(type.name);
        },
        gqlType: function gqlType(type) {
          var gqlType = pgGetGqlTypeByTypeIdAndModifier(type.id, type.attrTypeModifier);

          switch (gqlType.name) {
            case 'GeometryInterface':
            case 'GeometryPoint':
            case 'GeometryPolygon':
              return 'GeoJSON';

            default:
              return gqlType;
          }
        },
        subtype: function subtype(type) {
          var gqlType = pgGetGqlTypeByTypeIdAndModifier(type.id, type.attrTypeModifier);

          switch (gqlType.name) {
            case 'GeometryInterface':
            case 'GeometryPoint':
            case 'GeometryPolygon':
              return gqlType.name;

            default:
              return null;
          }
        },
        typmod: function typmod(type) {
          var modifier = type.attrTypeModifier;
          if (!modifier) return null;

          if (type.name === 'geography' || type.name === 'geometry') {
            // Ref: https://github.com/postgis/postgis/blob/2.5.2/liblwgeom/liblwgeom.h.in#L156-L173
            // #define TYPMOD_GET_SRID(typmod) ((((typmod) & 0x0FFFFF00) - ((typmod) & 0x10000000)) >> 8)
            // #define TYPMOD_GET_TYPE(typmod) ((typmod & 0x000000FC)>>2)
            // #define TYPMOD_GET_Z(typmod) ((typmod & 0x00000002)>>1)
            // #define TYPMOD_GET_M(typmod) (typmod & 0x00000001)
            var srid = (modifier & 0x0fffff00) - (modifier & 0x10000000) >> 8;
            var subtype = (modifier & 0x000000fc) >> 2;
            var hasZ = (modifier & 0x00000002) >> 1 === 1;
            var hasM = (modifier & 0x00000001) === 1;

            if (subtype < GIS_TYPES.length) {
              return {
                srid: srid,
                subtype: subtype,
                hasZ: hasZ,
                hasM: hasM,
                gisType: GIS_TYPES[subtype]
              };
            }
          }

          return {
            modifier: modifier
          };
        },
        modifier: function modifier(type) {
          return type.attrTypeModifier;
        },
        isArray: function isArray(type) {
          return type.isPgArray;
        }
      },
      MetaschemaField: {
        /** @param attr {import('graphile-build-pg').PgAttribute} */
        name: function name(attr) {
          return inflection.column(attr);
        },

        /** @param attr {import('graphile-build-pg').PgAttribute} */
        type: function type(attr) {
          if (attr.typeModifier > 0) {
            return _objectSpread(_objectSpread({}, attr.type), {}, {
              attrTypeModifier: attr.typeModifier
            });
          }

          return attr.type;
        }
      },
      MetaschemaTableInflection: {
        deleteByPrimaryKey: function deleteByPrimaryKey(table) {
          var _table$primaryKeyCons, _table$primaryKeyCons2;

          if (!((_table$primaryKeyCons = table.primaryKeyConstraint) === null || _table$primaryKeyCons === void 0 ? void 0 : (_table$primaryKeyCons2 = _table$primaryKeyCons.keyAttributes) === null || _table$primaryKeyCons2 === void 0 ? void 0 : _table$primaryKeyCons2.length)) return null;
          return inflection.deleteByKeys(table.primaryKeyConstraint.keyAttributes, table, table.primaryKeyConstraint);
        },
        updateByPrimaryKey: function updateByPrimaryKey(table) {
          var _table$primaryKeyCons3, _table$primaryKeyCons4;

          if (!((_table$primaryKeyCons3 = table.primaryKeyConstraint) === null || _table$primaryKeyCons3 === void 0 ? void 0 : (_table$primaryKeyCons4 = _table$primaryKeyCons3.keyAttributes) === null || _table$primaryKeyCons4 === void 0 ? void 0 : _table$primaryKeyCons4.length)) return null;
          return inflection.updateByKeys(table.primaryKeyConstraint.keyAttributes, table, table.primaryKeyConstraint);
        },
        createField: function createField(table) {
          return inflection.createField(table);
        },
        createInputType: function createInputType(table) {
          return inflection.createInputType(table);
        },
        allRows: function allRows(table) {
          return inflection.allRows(table);
        },
        allRowsSimple: function allRowsSimple(table) {
          return inflection.allRowsSimple(table);
        },
        tableFieldName: function tableFieldName(table) {
          return inflection.tableFieldName(table);
        },
        tableType: function tableType(table) {
          return inflection.tableType(table);
        },
        orderByType: function orderByType(table) {
          return inflection.orderByType(inflection.tableType(table));
        },
        filterType: function filterType(table) {
          if (typeof inflection.filterType === 'function') return inflection.filterType(inflection.tableType(table));
          return null;
        },
        inputType: function inputType(table) {
          return inflection.inputType(inflection.tableType(table));
        },
        patchType: function patchType(table) {
          return inflection.patchType(inflection.tableType(table));
        },
        conditionType: function conditionType(table) {
          return inflection.conditionType(inflection.tableType(table));
        },
        patchField: function patchField(table) {
          return inflection.patchField(inflection.tableType(table));
        },
        edge: function edge(table) {
          return inflection.edge(inflection.tableType(table));
        },
        edgeField: function edgeField(table) {
          return inflection.edgeField(table);
        },
        connection: function connection(table) {
          return inflection.connection(inflection.tableType(table));
        },
        typeName: function typeName(table) {
          return inflection._typeName(table);
        },
        enumType: function enumType(table) {
          return inflection.enumType(table);
        },
        createPayloadType: function createPayloadType(table) {
          return inflection.createPayloadType(table);
        },
        updatePayloadType: function updatePayloadType(table) {
          return inflection.updatePayloadType(table);
        },
        deletePayloadType: function deletePayloadType(table) {
          return inflection.deletePayloadType(table);
        }
      },
      MetaschemaTableQuery: {
        "delete": function _delete(table) {
          var _table$primaryKeyCons5, _table$primaryKeyCons6;

          if (!((_table$primaryKeyCons5 = table.primaryKeyConstraint) === null || _table$primaryKeyCons5 === void 0 ? void 0 : (_table$primaryKeyCons6 = _table$primaryKeyCons5.keyAttributes) === null || _table$primaryKeyCons6 === void 0 ? void 0 : _table$primaryKeyCons6.length)) return null;
          return inflection.deleteByKeys(table.primaryKeyConstraint.keyAttributes, table, table.primaryKeyConstraint);
        },
        update: function update(table) {
          var _table$primaryKeyCons7, _table$primaryKeyCons8;

          if (!((_table$primaryKeyCons7 = table.primaryKeyConstraint) === null || _table$primaryKeyCons7 === void 0 ? void 0 : (_table$primaryKeyCons8 = _table$primaryKeyCons7.keyAttributes) === null || _table$primaryKeyCons8 === void 0 ? void 0 : _table$primaryKeyCons8.length)) return null;
          return inflection.updateByKeys(table.primaryKeyConstraint.keyAttributes, table, table.primaryKeyConstraint);
        },
        create: function create(table) {
          return inflection.createField(table);
        },
        all: function all(table) {
          return inflection.allRows(table);
        },
        one: function one(table) {
          return inflection.tableFieldName(table);
        }
      },
      MetaschemaTable: {
        /** @param table {import('graphile-build-pg').PgClass} */
        name: function name(table) {
          return inflection.tableType(table); // return inflection._tableName(table);
        },

        /** @param table {import('graphile-build-pg').PgClass} */
        fields: function fields(table) {
          return table.attributes.filter(function (attr) {
            if (attr.num < 1) return false; // low-level props

            return true;
          });
        },

        /** @param table {import('graphile-build-pg').PgClass} */
        inflection: function inflection(table) {
          // return table so the MetaschemaTableInflection resolver uses that as input
          return table;
        },

        /** @param table {import('graphile-build-pg').PgClass} */
        query: function query(table) {
          return table;
        },

        /** @param table {import('graphile-build-pg').PgClass} */
        constraints: function constraints(table) {
          return table.constraints;
        },

        /** @param table {import('graphile-build-pg').PgClass} */
        foreignKeyConstraints: function foreignKeyConstraints(table) {
          return table.constraints.filter(function (c) {
            return c.type === 'f';
          });
        },

        /** @param table {import('graphile-build-pg').PgClass} */
        primaryKeyConstraints: function primaryKeyConstraints(table) {
          return table.constraints.filter(function (c) {
            return c.type === 'p';
          });
        },

        /** @param table {import('graphile-build-pg').PgClass} */
        uniqueConstraints: function uniqueConstraints(table) {
          return table.constraints.filter(function (c) {
            return c.type === 'u';
          });
        },

        /** @param table {import('graphile-build-pg').PgClass} */
        checkConstraints: function checkConstraints(table) {
          return table.constraints.filter(function (c) {
            return c.type === 'c';
          });
        },

        /** @param table {import('graphile-build-pg').PgClass} */
        exclusionConstraints: function exclusionConstraints(table) {
          return table.constraints.filter(function (c) {
            return c.type === 'x';
          });
        }
      },
      MetaschemaConstraint: {
        /** @param obj {import('graphile-build-pg').PgConstraint} */
        __resolveType: function __resolveType(obj) {
          switch (obj.type) {
            case 'p':
              return 'MetaschemaPrimaryKeyConstraint';

            case 'f':
              return 'MetaschemaForeignKeyConstraint';

            case 'c':
              return 'MetaschemaCheckConstraint';

            case 'u':
              return 'MetaschemaUniqueConstraint';

            case 'x':
              return 'MetaschemaExclusionConstraint';
          }
        }
      },
      Metaschema: {
        tables: function tables() {
          return introspection["class"].filter(function (kls) {
            if (!schemas.includes(kls.namespaceName)) return false; // r = ordinary table, i = index, S = sequence, t = TOAST table, v = view, m = materialized view, c = composite type, f = foreign table, p = partitioned table, I = partitioned index

            if (kls.classKind !== 'r') return false;
            return true;
          });
        }
      },
      Query: {
        _meta: function _meta() {
          // just placeholder
          return {};
        }
      }
    }
  };
});
exports.PgMetaschemaPlugin = PgMetaschemaPlugin;
var _default = PgMetaschemaPlugin;
exports["default"] = _default;